#
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#

package SnapMirrorAgent;
use NaServer;
use NaElement;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use ag_i18n_inc;

sub new {
	my ($self, $resname, $filername, $lang) = @_;

	my $class = ref($self) || $self;
	my $filer = {};

	$filer->{START_TIME} = $^T;
	$filer->{resname} = $resname;
	$filer->{LANG} = $lang;
	$filer->{FILER_NAME} = $filername;
	$filer->{SNAPMIRROR_ON}	= "false";
	$filer->{IGNORE_RELEASE} = 0;
	return bless $filer, $class;
}

sub OpenConnection {
	my ($self, $user, $enc_pword) = @_;

	my $filername = $self->{FILER_NAME};

	my $handle = NaServer->new($filername, 1, 1);

	$enc_pword =~s/^\s+//;
	$enc_pword =~s/\s+$//;

	# 
	# if password is specified, use it to login as root on the filer
	# if not, use the hosts.equiv file on filer to determine access rights.
	#
	if ($enc_pword) {
		my $Pword = `$HADECRYPT $enc_pword`; 
		if ($? != 0) {
			VCSAG_LOG_MSG("W", "Encountered errors while decrypting password!", 1002);
			$Pword = "";
		}
		$handle->set_admin_user($user, $Pword);
	} else {
		$handle->set_style("HOSTS");
	}

	my $output = $handle->invoke("system-get-version");
	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'system-get-version' failed on filer $filername. Error $output->{'results_errno'}: $r", 1000, "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $version = $output->child_get_string( "version" );
	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOGDBG_MSG(1, "ONTAP version on filer $filername is: $version", "$filername", "$version");
	$ENV{"LANG"}="C";

	$self->{FILER_HANDLE} = $handle;
	$self->IgnoreSnapMirrorRelease($version);
	return 0;
}

sub SnapMirrorGetInfo {
	my ($self, $location) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-get-status", 
		"location", $location);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-get-status' failed for volume $location on filer $filername. Error $output->{'results_errno'}: $r", 1001, "$location", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return undef;
        }

	$self->{SNAPMIRROR_ON} = $output->child_get_string("is-available");

	# get snapmirror status
	my $statuslist = $output->child_get("snapmirror-status");
	if (!defined($statuslist) || ($statuslist eq "")) {
		# no snapmirror status to report
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Volume $location on filer $filername is not a snapmirror replicated volume", 1002, "$location", "$filername");
		$ENV{"LANG"}="C";
		return undef;
	}

	# iterate through snapmirror status info list.
	# NOTE: THIS AGENT CURRENTLY SUPPORTS ONLY ONE DESTINATION PER 
	# SOURCE VOLUME
	my @infolist = $statuslist->children_get();
	my @sminfolist = ();
	foreach my $sm_info (@infolist) {
		my $info = {};
		$info->{SRC_LOC} = $sm_info->child_get_string("source-location");
		($info->{SRC_FLR}, $info->{SRC_VOL}) = split(/:/, $info->{SRC_LOC});
		$info->{DST_LOC} = $sm_info->child_get_string("destination-location");
		($info->{DST_FLR}, $info->{DST_VOL}) = split(/:/, $info->{DST_LOC});
		$info->{STATE} = $sm_info->child_get_string("state");
		$info->{STATUS} = $sm_info->child_get_string("status");
		$info->{BASE_SNAP} = $sm_info->child_get_string("base-snapshot");
		$info->{PROGRESS} = $sm_info->child_get_int("transfer-progress", 0);
		$info->{LAG} = $sm_info->child_get_int("lag-time", 0);
		push(@sminfolist, $info);

		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOGDBG_MSG(4, "Snapmirror info for volume $location on filer $filername: Source Location = $info->{SRC_LOC}, Destination Location = $info->{DST_LOC}, State = $info->{STATE}, Status = $info->{STATUS}, Base Snapshot = $info->{BASE_SNAP}, Progress = $info->{PROGRESS}, Lag = $info->{LAG}", "$location", "$filername", "$info->{SRC_LOC}", "$info->{DST_LOC}", "$info->{STATE}", "$info->{STATUS}", "$info->{BASE_SNAP}", "$info->{PROGRESS}", "$info->{LAG}");
		$ENV{"LANG"}="C";
	}
	#$self->{SNAPMIRROR_INFO} = \@sminfolist;

	#foreach my $sm_info (@{$self->{SNAPMIRROR_INFO}}) {
	#	for my $info (keys %{$sm_info}) {
	#		print "$info = $sm_info->{$info}\n";
	#	}
	#}
	return (@sminfolist);
}

sub SnapMirrorUpdate {
	my ($self, $dstloc, $srcloc, $wait) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	# wait for snapmirror status to be "idle"
	my $ret = $self->WaitForIdle($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Error while waiting for snapmirror status of volume $dstloc to become 'idle' on filer $filername", 1004, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $output = $handle->invoke("snapmirror-update", 
		"destination-location", $dstloc,
		"source-location", $srcloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-update' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1037, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("N", "snapmirror update initiated for volume $dstloc", 1028, "$dstloc");
	$ENV{"LANG"}="C";

	return 0 unless ($wait);

	# wait for update to complete or (TBD: timeout) 
	$ret = $self->WaitForIdle($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Error while waiting for snapmirror update of volume $dstloc to complete on filer $filername", 1005, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	return 0;
}

sub SnapMirrorQuiesce {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};
	my ( $output, $r );
	my $quiesce_retries = 3;
	my $quiesce_interval = 5;
	my $quiesce_failure = 0;

	
	#
	# wait for status to become idle or (TBD: timeout)
	#
	my $ret = $self->WaitForIdle($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$self->{LANG}";	
		VCSAG_LOG_MSG("E", "Error while waiting for snapmirror status of volume $dstloc to become 'idle' on filer $filername", 1004, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("N", "Attempting to quiesce snapmirror location $dstloc", 1043, "$dstloc");
	$ENV{"LANG"}="C";

        #
        # Quiesce fails in some cases where there are 
        # multiple volumes being simultaneously replicated. Retry
        #
	for ( my $ii = $quiesce_retries; $ii > 0; $ii-- ){
		$output = $handle->invoke("snapmirror-quiesce",
			"destination-location", $dstloc);

		if ($output->results_errno != 0) {
			$r = $output->results_reason();
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("W", "ONTAPI 'snapmirror-quiesce' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r. Retrying ...", 1044, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
			$ENV{"LANG"}="C";
			$quiesce_failure = 1;
		}else {
			$quiesce_failure = 0;
			last;
		}
		sleep $quiesce_interval ;
	}

	if ( $quiesce_failure == 1 ){
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-quiesce' failed $quiesce_retries times for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r. Giving up", 1046, "$quiesce_retries", "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}



        #
        # check if volume was successfully quiesced
        #
	my @sminfolist = $self->SnapMirrorGetInfo($dstloc);
	return 1 if (!@sminfolist);

	#
	# we expect to see only one entry in the snapmirror infolist.
	# this agent does not support multiple snapmirror destinations
	# for the same source volume
	#
	if (scalar(@sminfolist) != 1) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $filername returned multiple status entries. Administrative intervention required", 1003, "$dstloc", "$filername"); 
		$ENV{"LANG"}="C";
		return 1;
	}

	my $sm_info = $sminfolist[0];
	if ($sm_info->{STATE} ne "quiesced") {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "snapmirror-quiesce failed to quiesce the destination location $dstloc at filer $filername", 1045, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirrored location $dstloc has been successfully quiesced", 1047, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}



#
# Breaks a snapmirrored relationship on this filer for specified volume.
# Also queries status afterward to see if the volume was successfully
# broken-off
#
sub SnapMirrorBreak {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("N", "Attempting to break snapmirror location $dstloc", 1023, "$dstloc");
	$ENV{"LANG"}="C";

	my $output = $handle->invoke("snapmirror-break", 
		"destination-location", $dstloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-break' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1006, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	#
	# check if volume was successfully broken-off
	#
	my @sminfolist = $self->SnapMirrorGetInfo($dstloc);
	return 1 if (!@sminfolist);

	#
	# we expect to see only one entry in the snapmirror infolist.
	# this agent does not support multiple snapmirror destinations
	# for the same source volume
	#
	if (scalar(@sminfolist) != 1) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $filername returned multiple status entries. Administrative intervention required", 1003, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $sm_info = $sminfolist[0];
	if ($sm_info->{STATE} ne "broken-off") {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "snapmirror-break failed to break snapmirror relationship for destination location $dstloc at filer $filername", 1007, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}
	
	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirrored location $dstloc has been successfully broken-off", 1030, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}

#
#Get connection details of local filer and accordingly configure connection on remote filer 
#
sub SnapMirrorConfigureConnection {
	my ($self, $rfiler, $lfiler, $rflrname_attr, $srcloc) = @_;
	my ($srcflrname_conf, $volname) = split(/:/, $srcloc);
	if ($srcflrname_conf eq  $rflrname_attr) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("I", "SnapMirror over multiple paths is not configured ", 1101);
		$ENV{"LANG"}="C";
		return 0;
	} else {
		# srcflrname_conf is a connection name
		my ($ret, $name, $mode, $src1, $dst1, $src2, $dst2)=$lfiler->SnapMirrorGetConnection($srcflrname_conf);
		if ($ret == 100) {
			#
			#if remote filer allready has connection with same name but different details. Fail online process.
			#if remote filer has identical connection no need add new connection.
			#
			my ($retcode, $exst_name, $exst_mode, $exst_src1, $exst_dst1, $exst_src2, $exst_dst2) = $rfiler->SnapMirrorGetConnection($srcflrname_conf);
			if ($retcode == 100) { 
				if (($mode ne $exst_mode) || ($dst1 ne $exst_src1) || ($src1 ne $exst_dst1) || ($dst2 ne $exst_src2) || ($src2 ne $exst_dst2)) {
					$ENV{"LANG"}="$self->{LANG}";
                                	VCSAG_LOG_MSG("E", "Remote filer $rflrname_attr has existing connection $srcflrname_conf but with different connection details. Administrator action required", 1102, "$rflrname_attr", "$srcflrname_conf");
                                	$ENV{"LANG"}="C";
					return 1;
				} else {
					$ENV{"LANG"}="$self->{LANG}";
                                        VCSAG_LOG_MSG("I", "Remote filer $rflrname_attr has identical connection $srcflrname_conf, No need to add new connection", 1103, "$rflrname_attr", "$srcflrname_conf");
                                        $ENV{"LANG"}="C";
                                        return 100;
				}
			}
			#Set connection on Remote filer by swaping source and destination in address pair	
			if ($rfiler->SnapMirrorSetConnection($name, $mode, $dst1, $src1, $dst2, $src2) == 0) {
				$ENV{"LANG"}="$self->{LANG}";
				VCSAG_LOG_MSG("I", "Succesfully set SnapMirror connection $name on remote filer $rflrname_attr", 1104, "$name", "$rflrname_attr");
				$ENV{"LANG"}="C";
				return 100;
			} else {
				$ENV{"LANG"}="$self->{LANG}";
				VCSAG_LOG_MSG("E", "Failed to set SnapMirror connection $name on remote filer $rflrname_attr", 1105, "$name", "$rflrname_attr");
				$ENV{"LANG"}="C";
				return 1;
			} 
		} else {
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("E", "SnapMirror over multiple paths is configured but failed to get details of connection $srcflrname_conf", 1106, "$srcflrname_conf");
			$ENV{"LANG"}="C";
			return 1;
		}
	}
	return 0;
}

#
# Checks for connection information on this filer for specified connection.
# Return 0 if connection does not exists, 100 if exists.
#

sub SnapMirrorGetConnection {
	my ($self, $connection) = @_;

	my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};
	
	my $output = $handle->invoke ("snapmirror-list-connections",
                        "connection", $connection);
	
	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-list-connections' failed for connection $connection on filer $filername Error $output->{'results_errno'}: $r", 1107, "$connection", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
                return 1;
        }

	my $connlist = $output->child_get("snapmirror-connections");
        if (!defined($connlist) || ($connlist eq "")) {
                # no snapmirror connection to report
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "No snapmirror connection exists with name  $connection at filer $filername", 1108, "$connection", "$filername");
                $ENV{"LANG"}="C";
                return 1;
        } else {
		#
		# There should be only one connection (if it exists)
		#
		my @connections = $connlist->children_get();
		foreach my $conn_info(@connections) {
			my $name = $conn_info->child_get_string("name");
			my $mode = $conn_info->child_get_string("mode");

			my $pair1 = $conn_info->child_get("address-pair1");

			#If first address is not present then get the snapmirror error
			if ($pair1 eq undef) {
				my $error_info = $conn_info->child_get("snapmirror-error");
				my $error_no = $error_info->child_get_string("errno");
				my $error = $error_info->child_get_string("reason");

				$ENV{"LANG"}="$self->{LANG}";
                		VCSAG_LOG_MSG("E", "Failed to get details for connection name $connection, SnapMirror error $error_no :: $error", 1109, "$connection", "$error_no", "$error");
                		$ENV{"LANG"}="C";
				return 1;
			}
			my $dst1 = $pair1->child_get_string("destination-addr");
			my $src1 = $pair1->child_get_string("source-addr");
			
			#second address pair may or may not present
  			my $pair2 = $conn_info->child_get("address-pair2");
			my $src2 = "";
			my $dst2 = "";
			if ($pair2 ne undef) {
				$dst2 = $pair2->child_get_string("destination-addr");
                        	$src2 = $pair2->child_get_string("source-addr");
			}
			$ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("N", "Connection details Name :: $name, Mode :: $mode, Source1 :: $src1, Destination1 :: $dst1, Source2 :: $src2, Destination2 :: $dst2",1110, "$name", "$mode", "$src1", "$dst1", "$src2", "$dst2");
                        $ENV{"LANG"}="C";

			return (100, $name, $mode, $src1, $dst1, $src2, $dst2);
		}
	}

	return 0;
}

#
# Set connection information on filer for specified connection.
# Return 0 on success, 1 if fails.
#
sub SnapMirrorSetConnection {
	my ($self, $name, $mode, $src1, $dst1, $src2, $dst2) = @_;

	my $filername = $self->{FILER_NAME};
        my $handle = $self->{FILER_HANDLE};

	#form input xml request
	my $in = NaElement->new("snapmirror-set-connection");
	my $pair1 = NaElement->new("address-pair1");
	my $addr_pair1 = NaElement->new("address-pair");


	$addr_pair1->child_add_string("destination-addr", $dst1 );
	$addr_pair1->child_add_string("source-addr", $src1);
	$pair1->child_add($addr_pair1);
	$in->child_add($pair1);
	
	#second address pair may or may not present
	if (($src2 ne "") || ($dst2 ne "")) {
		my $pair2 = NaElement->new("address-pair2");
		my $addr_pair2 = NaElement->new("address-pair");
		$addr_pair2->child_add_string("destination-addr", $dst2);
		$addr_pair2->child_add_string("source-addr", $src2);
		$pair2->child_add($addr_pair2);
		$in->child_add($pair2);
	}

	$in->child_add_string("connection", $name);
	$in->child_add_string("mode", $mode);

	my $output = $handle->invoke_elem($in);

	if ($output->results_errno != 0) {
                my $r = $output->results_reason();
                $ENV{"LANG"}="$self->{LANG}";
                VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-set-connection' failed for connection $name  at filer $filername. Error $output->{'results_errno'}: $r", 1111, "$name", "$filername", "$output->{'results_errno'}", "$r");
                $ENV{"LANG"}="C";
                return 1;
        }

        $ENV{"LANG"}="$self->{LANG}";
        VCSAG_LOG_MSG("I", "snapmirror connection set with  $name", 1031, "$name");
        $ENV{"LANG"}="C";

        return 0;
}


#
# Checks for schedule information on this filer for specified volume.
# Return 0 if schedule does not exists, 100 if exists.
#
sub SnapMirrorGetSchedule {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-list-schedule",	
		"destination-location", $dstloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-list-schedule' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1008, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $schedlist = $output->child_get("snapmirror-schedules");
	if (!defined($schedlist) || ($schedlist eq "")) {
		# no snapmirror schedule to report
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("N", "No snapmirror schedule exists for location $dstloc at filer $filername", 1009, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	} else {

		#
		# there should be only 1 schedule (if it exists)
		#
		my @schedules = $schedlist->children_get();
		foreach my $sm_info (@schedules) {
                        my $hours = $sm_info->child_get_string("hours");
                        my $minutes = $sm_info->child_get_string("minutes");
                        $ENV{"LANG"}="$self->{LANG}";
                        VCSAG_LOG_MSG("N", "$hours and $minutes",1039, "$hours", "$minutes");
                        $ENV{"LANG"}="C";
			my $srcloc = $sm_info->child_get_string("source-location");
			return (100, $srcloc);
                }

	}

	return 0;
}

#
# Checks for schedule information on this filer for specified volume.
# Return 0 if schedule does not exists, 100 if exists.
#
sub SnapMirrorGetSyncSchedule {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-list-sync-schedule",
			"destination-location", $dstloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-list-sync-schedule' failed for volume $dstloc on filer $filername. Error $output->{'results_errno'}: $r", 1040, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $schedlist = $output->child_get("snapmirror-sync-schedules");
	if (!defined($schedlist) || ($schedlist eq "")) {
		# no snapmirror schedule to report
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("N", "No snapmirror schedule exists for location $dstloc at filer $filername", 1009, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	} else {

		#
		# there should be only 1 schedule (if it exists)
		#
		my @schedules = $schedlist->children_get();
		foreach my $sm_info (@schedules) {
			my $ops_throttle = $sm_info->child_get_string("ops-throttle");
			my $visibility = $sm_info->child_get_string("visibility-frequency");
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("N", "$ops_throttle and $visibility", 1050, "$ops_throttle", "$visibility");
			$ENV{"LANG"}="C";
			my $srcloc = $sm_info->child_get_string("source-location");
			return (100, $srcloc);
		}
	}

	return 0;
}

sub SnapMirrorSetSchedule {
	my ($self, $dstloc, $srcloc, $sched) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my ($minute, $hour, $dayofmonth, $dayofweek) = split(/\s+/, $sched);
	my $output = $handle->invoke("snapmirror-set-schedule", 
		"days-of-month", $dayofmonth,
		"days-of-week", $dayofweek,
		"destination-location", $dstloc,
		"hours", $hour,
		"minutes", $minute,
		"source-location", $srcloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-set-schedule' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1010, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirror schedule set for location $dstloc", 1031, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}

sub SnapMirrorSetSyncSchedule {
	my ($self, $dstloc, $srcloc, $ops_throttle, $visibility) = @_;
	
	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};
	my $output;
	if(!defined($ops_throttle) || ($ops_throttle eq "")){
	$output = $handle->invoke("snapmirror-set-sync-schedule",
		"destination-location", $dstloc,
		"source-location", $srcloc,
		"visibility-frequency", $visibility);
	}
	else {
	$output = $handle->invoke("snapmirror-set-sync-schedule",
		"destination-location", $dstloc,
		"ops-throttle", $ops_throttle,
		"source-location", $srcloc,
		"visibility-frequency", $visibility);
	}
	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-set-sync-schedule' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1041, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}

	
	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirror schedule set for location $dstloc", 1031, "$dstloc");
	$ENV{"LANG"}="C";

        return 0;
}

sub SnapMirrorDeleteSchedule {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-delete-schedule", 
		"destination-location", $dstloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-delete-schedule' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1011, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirror schedule deleted for location $dstloc", 1032, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}

sub SnapMirrorDeleteSyncSchedule {
	my ($self, $dstloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-delete-sync-schedule",
		"destination-location", $dstloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-delete-sync-schedule' failed for destination location $dstloc at filer $filername. Error $output->{'results_errno'}: $r", 1042, "$dstloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirror schedule deleted for location $dstloc", 1032, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}


	

#
# This routine intiates a snapmirror resync. It is up to the caller 
# to wait for resync to complete, if required.
#
sub SnapMirrorResync {
	my ($self, $dstloc, $srcloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};
	my ($output, $r);

	my $resync_retries = 50;
	my $resync_interval = 5;
	my $resync_failure = 0;


	#
	# Resync fails in some cases. 
	# Wait for some time and retry again.
	#
	for (my $ii = $resync_retries; $ii > 0; $ii--){
		$output = $handle->invoke("snapmirror-resync", "destination-location",
			$dstloc, "source-location", $srcloc);
		if ($output->results_errno != 0) {
			$r = $output->results_reason();
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("I", "ONTAPI 'snapmirror-resync' failed for destination location $dstloc, source location $srcloc at filer $filername. Error $output->{'results_errno'}: $r. Retrying ...", 1012, "$dstloc", "$srcloc", "$filername", "$output->{'results_errno'}", "$r");
			$ENV{"LANG"}="C";
			$resync_failure = 1;
			sleep $resync_interval;
		}else{
			$resync_failure = 0;
			last;
		}
	}
	
	if ($resync_failure == 1) {
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-resync' failed for destination location $dstloc, source location $srcloc at filer $filername. Error $output->{'results_errno'}: $r. Giving up ...", 1054, "$dstloc", "$srcloc", "$filername", "$output->{'results_errno'}", "$r");
			$ENV{"LANG"}="C";
			return 1;
	}


	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("N", "snapmirror resync initiated for destination location $dstloc", 1029, "$dstloc");
	$ENV{"LANG"}="C";

	return 0;
}

#
# Release the previous snapmirror relationship on this filer.
# i,e inform snapmirror on this filer (i.e. original source location) that a
# direct mirror (i.e. original destination location) is no longer going to make
# snapmirror requests.
#
sub SnapMirrorRelease {
	my ($self, $dstloc, $srcloc) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	my $output = $handle->invoke("snapmirror-release", 
		"destination-location", $dstloc,
		"source-location", $srcloc);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-release' failed for destination location $dstloc, source location $srcloc at filer $filername. Error $output->{'results_errno'}: $r", 1013, "$dstloc", "$srcloc", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "snapmirror release succeeded for destination location $dstloc, source location $srcloc", 1034, "$dstloc", "$srcloc");
	$ENV{"LANG"}="C";

	return 0;
}

sub SnapshotDelete {
	my ($self, $volname, $ssname) = @_;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};
	
	#Find volumename if qtree is configured.
	$volname =~ /vol\/(.*?)\//;
        if ($1 ne "") {
                VCSAG_LOG_MSG("I", "Qtree Snapmirror is configured. Qtree = $volname and Volume = $1", 2000, "$volname", "$1");
                $volname = $1;
        }	

	my $output = $handle->invoke("snapshot-delete", 
		"volume", $volname,
		"snapshot", $ssname);

	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapshot-delete' failed for snapshot $ssname on volume $volname at filer $filername. Error $output->{'results_errno'}: $r", 1014, "$ssname", "$volname", "$filername", "$output->{'results_errno'}", "$r");
		$ENV{"LANG"}="C";
		return 1;
        }

	$ENV{"LANG"}="$self->{LANG}";
	VCSAG_LOG_MSG("I", "Base snapshot $ssname deleted for volume $volname at filer $filername", 1035, "$ssname", "$volname", "$filername");
	$ENV{"LANG"}="C";

	return 0;
}

#
# This routine will wait in a tight loop for the snapmirror status
# of the specified volume to become 'idle'. This is required because
# certain snapmirror commands can only be issued when the status is
# idle or else the command will fail.
#
sub WaitForIdle {
	my ($self, $location) = @_;

	my @sminfolist;

	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};

	# wait for snapmirror status to be "idle"
	while (1) {
		@sminfolist = $self->SnapMirrorGetInfo($location);
		return 1 if (!@sminfolist);

		#
		# we expect to see only one entry in the snapmirror infolist.
                # this agent does not support multiple snapmirror destinations
                # for the same source volume
		#
		if (scalar(@sminfolist) != 1) {
			$ENV{"LANG"}="$self->{LANG}";
			VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $location on filer $filername returned multiple status entries. Administrative intervention required", 1003, "$location", "$filername");
			$ENV{"LANG"}="C";
			return 1;
		}
	
		my $sm_info = $sminfolist[0];
		last if ($sm_info->{STATUS} eq "idle" || $sm_info->{STATUS} eq "in-sync");
		undef(@sminfolist);	
		# sleep 5 seconds before checking again
		sleep 5;
 
	}

	return 0;
}

sub get_state {
	my ($self, $location) = @_;

	my $filername = $self->{FILER_NAME};
	my @sminfolist = $self->SnapMirrorGetInfo($location);
	return undef if (!@sminfolist);

	#
	# we expect to see only one entry in the snapmirror infolist.
	# this agent does not support multiple snapmirror destinations
	# for the same source volume
	#
	if (scalar(@sminfolist) > 1) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $location on filer $filername returned multiple status entries. Administrative intervention required", 1003, "$location", "$filername");
		$ENV{"LANG"}="C";
		return undef;
	}

	return $sminfolist[0];
}

sub failover_required {
	my ($self, $sm_info) = @_;  
	
	return 0 if (($sm_info->{STATE} eq "source") || 
		     ($sm_info->{STATE} eq "broken-off"));
	return 1 if ($sm_info->{STATE} eq "snapmirrored");

	return -1;
}

sub failover_time {
}

sub takeover {
	my ($self, $sm_info) = @_;
	
	my $filername = $self->{FILER_NAME};
	my $dstloc = $sm_info->{DST_LOC};
	
	#
	# check if snapmirror state is 'snapmirrored'
	#
	my @sminfolist = $self->SnapMirrorGetInfo($dstloc);
	return 1 if (!@sminfolist);

	#
	# we expect to see only one entry in the snapmirror infolist.
	# this agent does not support multiple snapmirror destinations
	# for the same source volume
	#
	if (scalar(@sminfolist) != 1) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("I", "Multiple snapmirror destinations for a volume is not supported by this agent. 'snapmirror status' for volume $dstloc on filer $filername returned multiple status entries. Administrative intervention required", 1003, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $new_sm_info = $sminfolist[0];
	return 0 if ($new_sm_info->{STATE} eq "broken-off");

	if ($new_sm_info->{STATE} ne "snapmirrored") {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Volume $dstloc is in state $new_sm_info->{STATE}", 1021, "$dstloc", "$new_sm_info->{STATE}");
		$ENV{"LANG"}="C";
		return 1;
	}

	my $ret = $self->SnapMirrorQuiesce($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to queisce destination location $dstloc on filer $filername", 1048, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ret = $self->SnapMirrorBreak($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to break destination location $dstloc on filer $filername", 1051, "$dstloc", "$filername");
		$ENV{"LANG"}="C";
		return 1;
	}

	#	
	# if schedule exists on local filer, delete it.
	# SnapMirrorGetSchedule() return 100 if schedule exists and
	# 0 if schedule does not exist for specified location.
	#
	#my $sched_exist = $self->SnapMirrorGetSchedule($dstloc);
	#if ($sched_exists == 100) {
	#	$ret = $self->SnapMirrorDeleteSchedule($dstloc);
	#	return 1 if ($ret);
	#} elsif ($sched_exists == 1) {
	#	# log error
	#	return 1;
	#}
	
	return 0;	
}

sub failback {
}

sub migrate {
	my ($lfiler, $rfiler, $sm_info, $schedule, $ops_throttle, $visibility) = @_;
	
	#
	# volume on lfiler filer is the current destination (to be made source)
	# volume on remote filer is the current source (to be made destination)
	#
	my $dstloc = $sm_info->{DST_LOC};
	my $srcloc = $sm_info->{SRC_LOC};

	my ($sched_exists, $source);
	my $srcloc_to_config;
	my $dstloc_to_config;

	my ($rflrname_conf, $rvolname_conf) = split(/:/, $srcloc);
	my ($lflrname, $lvolname) = split(/:/, $dstloc);
	my $rflrname = $rfiler->{FILER_NAME};

	my $ret = $lfiler->SnapMirrorUpdate($dstloc, $srcloc, 1);
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to update snapmirror location $dstloc on filer $lflrname", 1020, "$dstloc", "$lflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ret = $lfiler->SnapMirrorQuiesce($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to queisce destination location $dstloc on filer $lflrname", 1048, "$dstloc", "$lflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	$ret = $lfiler->SnapMirrorBreak($dstloc);
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to break destination location $dstloc on filer $lflrname", 1051, "$dstloc", "$lflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	#
	# delete schedule on local filer, if it exists 
	# Configure connection if required.
	#
	($sched_exists, $source) = $lfiler->SnapMirrorGetSyncSchedule($dstloc);
	if ($sched_exists == 100) {
		my $retcode = $lfiler->SnapMirrorConfigureConnection($rfiler, $lfiler, $rflrname, $srcloc);
                if ($retcode == 1) {
                 	$ENV{"LANG"}="$lfiler->{LANG}";
                        VCSAG_LOG_MSG("E", "Failed to configure connection  $source on filer $rflrname", 1100, "$source", "$rflrname");
                        $ENV{"LANG"}="C";
                        return 1;
                }
		$ret = $lfiler->SnapMirrorDeleteSyncSchedule($dstloc);
		if ($retcode = 100) {
                	$srcloc_to_config = $rflrname.":".$rvolname_conf;
                        $dstloc_to_config = $rflrname_conf.":".$lvolname;
                } else {
			$srcloc_to_config = $srcloc;
                        $dstloc_to_config = $dstloc;
                }
		
	} else {
		($sched_exists, $source) = $lfiler->SnapMirrorGetSchedule($dstloc);
		if ($sched_exists == 100) {
			my $retcode = $lfiler->SnapMirrorConfigureConnection($rfiler, $lfiler, $rflrname, $srcloc);
			if ($retcode == 1) {
				$ENV{"LANG"}="$lfiler->{LANG}";
                		VCSAG_LOG_MSG("E", "Failed to configure connection  $source on filer $rflrname", 1100, "$source", "$rflrname");
                		$ENV{"LANG"}="C";
				return 1;
			}
			$ret = $lfiler->SnapMirrorDeleteSchedule($dstloc);
			if ($retcode == 100) {
                                $srcloc_to_config = $rflrname.":".$rvolname_conf;
                                $dstloc_to_config = $rflrname_conf.":".$lvolname;
                        } else {
				$srcloc_to_config = $srcloc;
				$dstloc_to_config = $dstloc;
			}
		}
	}	
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("W", "Failed to delete snapmirror schedule for location $dstloc on filer $lflrname", 1022, "$dstloc", "$lflrname");
		$ENV{"LANG"}="C";
		#return 1;
	}

	#
	# Add new schedule on remote filer.
	# What used to be the source is now going to be made a destination
	#

	if(!defined($schedule) || ($schedule eq "")){
	$ret = $rfiler->SnapMirrorSetSyncSchedule($srcloc_to_config, $dstloc_to_config, $ops_throttle, $visibility);
	}
	else {
	$ret = $rfiler->SnapMirrorSetSchedule($srcloc_to_config, $dstloc_to_config, $schedule);
	}

	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to set snapmirror schedule for new destination location $srcloc_to_config on filer $rflrname", 1024, "$srcloc_to_config", "$rflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	#
	# Resync new destination (i.e remote filer) with new 
	# source (i.e local filer)
	#
	$ret = $rfiler->SnapMirrorResync($srcloc_to_config, $dstloc_to_config);
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to resync new destination location $srcloc_to_config on filer $rflrname", 1038, "$srcloc_to_config", "$rflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	# 
	# wait for resync to complete
	#
	my $done = 0;
	while (!$done) {
		my @sminfolist = $rfiler->SnapMirrorGetInfo($srcloc_to_config);
		return 1 if (!@sminfolist);

		#
		# find the right snapmirror-info entry in the infolist and 
		# wait till Status goes to Idle (there should ideally be only 
		# 2 entries in the list) 
		# TBD: we need to consider situations where status goes to 
		# something other that 'idle' after a resync is initiated.
		#
		foreach my $sm_info (@sminfolist) {
			next if ($sm_info->{DST_LOC} ne $srcloc_to_config);

			if ($sm_info->{STATUS} eq "idle" || $sm_info->{STATUS} eq "in-sync") {
				$done = 1;
			} else {
				sleep 5;
				$ENV{"LANG"}="$lfiler->{LANG}";
				VCSAG_LOG_MSG("N", "Waiting for snapmirror resync to destination location $srcloc on filer $rflrname to complete", 1025, "$srcloc", "$rflrname");
				$ENV{"LANG"}="C";
			}
			last;
		}
	}

	#
	# Find and delete base-snapshot from local filer (i.e original dest)
	#
	my @sminfolist = $lfiler->SnapMirrorGetInfo($dstloc);
	return 1 if (!@sminfolist);

	#
	# find the right snapmirror-info entry in the infolist and get the
	# base snapshot name (there should ideally be only 2 entries in the 
	# list). 
	# This extra processing is required to work around the problematic 
	# status output from 'snapmirror-get-status' after a resync is issued 
	# to reverse direction of snapmirror replication.
	#
	$done = 0;
	foreach my $sm_info (@sminfolist) {
		if ($sm_info->{DST_LOC} eq $dstloc) {
			$ret = $lfiler->SnapshotDelete($lvolname, 
						$sm_info->{BASE_SNAP});
			if ($ret) {
				$ENV{"LANG"}="$lfiler->{LANG}";
				VCSAG_LOG_MSG("E", "Failed to delete base snapshot $sm_info->{BASE_SNAP} for volume $lvolname on filer $lflrname", 1033, "$sm_info->{BASE_SNAP}", "$lvolname", "$lflrname");
				$ENV{"LANG"}="C";
				return 1;	
			}
			$done = 1;
			last;
		}
	}

	if (!$done) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to find base snapshot for volume $lvolname on filer $lflrname", 1026, "$lvolname", "$lflrname");
		$ENV{"LANG"}="C";
		return 1;	
	}

	#
	# Release previous snapmirror relationship at 
	# remote filer (i.e. original source)	
	#
	$ret = $rfiler->SnapMirrorRelease($dstloc, $srcloc);
	if ($ret) {
		$ENV{"LANG"}="$lfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to release previous snapmirror relationship between original source location $srcloc and original destination location $dstloc on filer $rflrname", 1027, "$srcloc", "$dstloc", "$rflrname");
		$ENV{"LANG"}="C";
		return 1 if (!$rfiler->{IGNORE_RELEASE});
                VCSAG_LOG_MSG("W", "IGNORE_RELEASE flag is set. Ignore snapmirror-release API.", 1058);

	}

	return 0;
}

sub swap_role {
	my ($org_dstfiler, $org_srcfiler, $org_dstloc, $org_srcloc, $schedule, $ops_throttle, $visibility) = @_;
	my $ret;
	my ($sched_exists, $source);
	my $source;	
	#
	# volume on org_dstfiler filer is to be made the new source
	# volume on org_srcfiler filer is to be made the new destination
	#

	my ($org_srcflrname, $org_srcvol) = split(/:/, $org_srcloc);
	my ($org_dstflrname, $org_dstvol) = split(/:/, $org_dstloc);
	my $srcloc_to_config;
	my $dstloc_to_config;
	#
	# delete schedule on original destination filer, if it exists
	#
	($sched_exists, $source) = $org_dstfiler->SnapMirrorGetSyncSchedule($org_dstloc);
	my ($connection_name, $volname) = split(/:/, $source);
	if ($sched_exists == 100) {
		my $retcode = $org_dstfiler->SnapMirrorConfigureConnection($org_srcfiler, $org_dstfiler, $org_srcflrname, $source);
		#if ( $org_dstfiler->SnapMirrorConfigureConnection($org_srcfiler, $org_dstfiler, $org_srcflrname, $source) != 0) {
                if ($retcode == 1) {
	                $ENV{"LANG"}="$org_dstfiler->{LANG}";
                        VCSAG_LOG_MSG("E", "Failed to configure connection $org_srcloc on filer $org_srcflrname", 1100, "$source", "$org_srcflrname");
                        $ENV{"LANG"}="C";
                        return 1;
                }
		$ret = $org_dstfiler->SnapMirrorDeleteSyncSchedule($org_dstloc);
		if ($retcode == 100) {
                        $srcloc_to_config = $org_srcflrname.":".$org_srcvol;
                        $dstloc_to_config = $connection_name.":".$org_dstvol;
                } else {
                        $srcloc_to_config = $org_srcloc;
                        $dstloc_to_config = $org_dstloc;
                }

	}	
	else {
		($sched_exists, $source) = $org_dstfiler->SnapMirrorGetSchedule($org_dstloc);
		($connection_name, $volname) = split(/:/, $source);
		if ($sched_exists == 100) {
			my $retcode = $org_dstfiler->SnapMirrorConfigureConnection($org_srcfiler, $org_dstfiler, $org_srcflrname, $source);
                	if ($retcode == 1) {
                	        $ENV{"LANG"}="$org_dstfiler->{LANG}";
                     		VCSAG_LOG_MSG("E", "Failed to configure connection $org_srcloc on filer $org_srcflrname", 1100, "$source", "$org_srcflrname");
          		        $ENV{"LANG"}="C";
                        	return 1;
                	}
			$ret = $org_dstfiler->SnapMirrorDeleteSchedule($org_dstloc);
			if ($retcode == 100) {
                 	       $srcloc_to_config = $org_srcflrname.":".$org_srcvol;
                 	       $dstloc_to_config = $connection_name.":".$org_dstvol;
                	} else {
                 	       $srcloc_to_config = $org_srcloc;
                  	       $dstloc_to_config = $org_dstloc;
               	        }
		}
	}
	if ($ret) {
		$ENV{"LANG"}="$org_dstfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to delete snapmirror schedule for location $org_dstloc on filer $org_dstflrname", 1022, "$org_dstloc", "$org_dstflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	#
	# Add new schedule on new destination filer.
	# What used to be the source is now going to be made a destination
	#
	
	if(!defined($schedule) || ($schedule eq "")){
	$ret = $org_srcfiler->SnapMirrorSetSyncSchedule($srcloc_to_config, $dstloc_to_config, $ops_throttle, $visibility);
	}
	else {
	$ret = $org_srcfiler->SnapMirrorSetSchedule($srcloc_to_config, $dstloc_to_config, $schedule);
	}
	if ($ret) {
		$ENV{"LANG"}="$org_dstfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to set snapmirror schedule for new destination location $srcloc_to_config on filer $org_srcflrname", 1024, "$srcloc_to_config", "$org_srcflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	#
	# Resync new destination with new source 
	#
	$ret = $org_srcfiler->SnapMirrorResync($srcloc_to_config, $dstloc_to_config);
	if ($ret) {
		$ENV{"LANG"}="$org_dstfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to resync new destination location $srcloc_to_config on filer $org_srcflrname", 1036, "$srcloc_to_config", "$org_srcflrname");
		$ENV{"LANG"}="C";
		return 1;
	}

	# 
	# wait for resync to complete
	#
	my $done = 0;
	while (!$done) {
		my @sminfolist = $org_srcfiler->SnapMirrorGetInfo($srcloc_to_config);
		return 1 if (!@sminfolist);

		#
		# find the right snapmirror-info entry in the infolist and 
		# wait till Status goes to Idle (there should ideally be only 
		# 2 entries in the list) 
		# TBD: we need to consider situations where status goes to 
		# something other that 'idle' after a resync is initiated.
		#
		foreach my $sm_info (@sminfolist) {
			next if ($sm_info->{DST_LOC} ne $srcloc_to_config);

			if ($sm_info->{STATUS} eq "idle" || $sm_info->{STATUS} eq "in-sync") {
				$done = 1;
			} else {
				sleep 5;
				$ENV{"LANG"}="$org_dstfiler->{LANG}";
				VCSAG_LOG_MSG("N", "Waiting for snapmirror resync to destination location $srcloc_to_config on filer $org_srcflrname to complete", 1025, "$org_srcloc", "$org_srcflrname");
				$ENV{"LANG"}="C";
			}
			last;
		}
	}

	#
	# Find and delete base-snapshot from original destination filer
	#
	my @sminfolist = $org_dstfiler->SnapMirrorGetInfo($org_dstloc);
	return 1 if (!@sminfolist);

	#
	# find the right snapmirror-info entry in the infolist and get the
	# base snapshot name (there should ideally be only 2 entries in the 
	# list). 
	# This extra processing is required to work around the problematic 
	# status output from 'snapmirror-get-status' after a resync is issued 
	# to reverse direction of snapmirror replication.
	#
	$done = 0;
	foreach my $sm_info (@sminfolist) {
		if ($sm_info->{DST_LOC} eq $org_dstloc) {
			$ret = $org_dstfiler->SnapshotDelete($org_dstvol, 
						$sm_info->{BASE_SNAP});
			if ($ret) {
				$ENV{"LANG"}="$org_dstfiler->{LANG}";
				VCSAG_LOG_MSG("E", "Failed to delete base snapshot $sm_info->{BASE_SNAP} for volume $org_dstvol on filer $org_dstflrname", 1033, "$sm_info->{BASE_SNAP}", "$org_dstvol", "$org_dstflrname");
				$ENV{"LANG"}="C";
				return 1;	
			}
			$done = 1;
			last;
		}
	}

	if (!$done) {
		$ENV{"LANG"}="$org_dstfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to find base snapshot for volume $org_dstvol on filer $org_dstflrname", 1026, "$org_dstvol", "$org_dstflrname");
		$ENV{"LANG"}="C";
		return 1;	
	}

	#
	# Release previous snapmirror relationship at original source filer
	#
	$ret = $org_srcfiler->SnapMirrorRelease($org_dstloc, $org_srcloc);
	if ($ret) {
		$ENV{"LANG"}="$org_dstfiler->{LANG}";
		VCSAG_LOG_MSG("E", "Failed to release previous snapmirror relationship between original source location $org_srcloc and original destination location $org_dstloc on filer $org_srcflrname", 1027, "$org_srcloc", "$org_dstloc", "$org_srcflrname");
		$ENV{"LANG"}="C";

                return 1 if (!$org_srcfiler->{IGNORE_RELEASE});
                VCSAG_LOG_MSG("W", "IGNORE_RELEASE flag is set. Ignore snapmirror-release API.", 1058);
	}

	return 0;
}

sub RepTechRPO {
	VCSAG_LOGDBG_MSG(3, "RepTechRPO START");
	my ($self, $VolName, $lfiler, $rfiler) = @_;
	my ($lag, $last_mirr, $RPO, $output, $snapMirrStatus, @snapinfo);
	my $filername = $self->{FILER_NAME};
	my $handle = $self->{FILER_HANDLE};
	
	$output = $handle->invoke("snapmirror-get-status","location", $VolName);
	if ($output->results_errno != 0) {
		my $r = $output->results_reason();
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "ONTAPI 'snapmirror-get-status' failed for volume $VolName on filer $filername. Error $output->{'results_errno'}: $r", 1052, "$VolName", "$filername", "$output->{'results_errno'}", "$r");
		print "ONTAPI 'snapmirror-get-status' failed for volume $VolName on filer $filername with Error $output->{'results_errno'}: $r\n";
		$ENV{"LANG"}="C";
		return 1;
	}

	$snapMirrStatus = $output->child_get("snapmirror-status");
	if (!defined($snapMirrStatus) || ($snapMirrStatus eq "")) {
# no snapmirror status to report
		$ENV{"LANG"}="$self->{LANG}";
		VCSAG_LOG_MSG("E", "Volume $VolName on filer $filername is not a snapmirror replicated volume", 1053, "$VolName", "$filername");
		print "Volume $VolName on filer $filername is not a snapmirror replicated volume\n";
		$ENV{"LANG"}="C";
		return 1;
	}


	@snapinfo = $snapMirrStatus->children_get();
	$RPO = 0;
	foreach my $snap_info_list (@snapinfo) {
		my $info = {};
		my $snap_state = $snap_info_list->child_get_string("state");
		if ($snap_state eq "snapmirrored" || $snap_state eq "quiesced") {
			$info->{SRC_LOC} = $snap_info_list->child_get_string("source-location");
			($info->{SRC_FLR}, $info->{SRC_VOL}) = split(/:/, $info->{SRC_LOC});
			$info->{DST_LOC} = $snap_info_list->child_get_string("destination-location");
			($info->{DST_LOC}, $info->{DST_VOL}) = split(/:/, $info->{DST_LOC}) if ($info->{DST_LOC} =~ /(\S+):(\S+)/);
			if ($info->{SRC_FLR} eq $rfiler && $info->{DST_LOC} eq $lfiler) {
				$lag = $snapinfo[0]->child_get_int("lag-time", 0);
				$lag = time_format($lag);
				$RPO = 1;
				VCSAG_LOG_MSG("I", "Replication RPO = $lag [hh:mm:ss] [As reported by NetApp SnapMirror].", 1054, "$lag");
				print "Replication RPO = $lag [hh:mm:ss] [As reported by NetApp SnapMirror].\n";
				VCSAG_LOGDBG_MSG(3, "RepTechRPO Successfull");
				return 0;
			} else {
				VCSAG_LOG_MSG ("E","Not able to report RPO because there is a mismatch in the configuration. Please verify that the attributes RemoteFilerName [$rfiler] and LocalFilerName [$lfiler] have values as per the configuration at the filer.", 1055, $rfiler, $lfiler);
				print "Not able to report RPO because there is a mismatch in the configuration. Please verify that the attributes RemoteFilerName [$rfiler] and LocalFilerName [$lfiler] have values as per the configuration at the filer.";
				return 1;
				VCSAG_LOGDBG_MSG(3, "RepTechRPO UnSuccessfull");
			}
		}
	}
	if (!$RPO) {
		VCSAG_LOG_MSG ("W", "Not able to report RPO because this action entry point should be invoked at the secondary site.\n", 1056);
		print "Not able to report RPO because this action entry point should be invoked at the secondary site.\n";
		VCSAG_LOGDBG_MSG(3, "RepTechRPO UnSuccessfull");
		return 1;
	}
}

sub time_format {
		my $time = shift;
		my ($hour, $min, $sec,$t1);
		$hour = "0";
		$min = "0";
		$sec = "0";
		if ($time >= 60){
				$t1 = int ($time / 60);
				$sec = ($time % 60);
				if ($t1 >= 60){
						$hour = int ($t1/60);
						$min = $t1 % 60;
				}else{
						$min = ($t1);
				}
		}
		else {
				$sec = $time;
		}
		$hour = "0$hour" if ($hour < 10 && $hour >= 0);
		$min = "0$min" if ($min < 10 && $min >= 0);
		$sec = "0$sec" if ($sec < 10 && $sec >= 0);
		return "$hour:$min:$sec";
}



sub IgnoreSnapMirrorRelease {

	my ($self, $version) = @_;
	my $sMajor = "";
	my $sMinor = "";
	VCSAG_LOGDBG_MSG(1, "$version");

	if ($version =~ /NetApp.*?(\d+)\.(\d+).*7-Mode/) {
		$sMajor = $1;
		$sMinor = $2;
	}

	VCSAG_LOGDBG_MSG(1, "Major version: $sMajor, Minor Version: $sMinor");

	if($sMajor > 8) {
		$self->{IGNORE_RELEASE} = 1;
		VCSAG_LOG_MSG("W", "IGNORE_RELEASE flag set. Snapmirror Release operation failure will be ignored.", 1057);
		return 0;
	}

        if(($sMajor == 8 ) && ( $sMinor >= 2)) {
                $self->{IGNORE_RELEASE} = 1;
                VCSAG_LOG_MSG("W", "IGNORE_RELEASE flag set. Snapmirror Release operation failure will be ignored.", 1058);
                return 0;
        }

	return 1;
}


1;
